"""_summary_


"""

# Internal dependencies (should ship with python)
import webbrowser                               # Used to open a browser tab
from tempfile import TemporaryDirectory
import os

# third party dependencies (need to be installed via pip)
from livereload import Server                   # Used to livereload pages in browser
from ezcv.core import generate_site             # Used to generate the static html pages
from flask import render_template, Flask        # Used to setup a running wsgi/http server
from jinja2.exceptions import TemplateNotFound  # Used to except template errors raised when a page doesn't exist

# Setup global variables
PORT = 5000 # The port to open flask to

def open_in_browser():
    """Opens localhost:{port} in whichever browser is installed"""
    browser_types = ["chromium-browser", "chromium", "chrome", "google-chrome", "firefox", "mozilla", "opera", "safari"] # A list of all the types of browsers to try
    for browser_name in browser_types:   # Look for which browser is installed
        try:
            webbrowser.get(browser_name) # Search for browser
            break                        # Browser has been found
        except webbrowser.Error:         # Browser wasn't found
            continue
    webbrowser.open(f"http://localhost:{PORT}", new=2) # Open the preview in the browser


def start_server(project_folder: str="."):
    """Preview entrypoint

    Parameters
    ----------
    project_folder : str, optional
        The folder where the config.yml and content folders are, by default "."
    """
    project_folder = os.path.abspath(project_folder)
    with TemporaryDirectory() as temp_dir:
        ## Configure the flask app
        app = Flask(__name__, static_url_path='', template_folder=temp_dir, static_folder=temp_dir)
        ### These are a security risk, if you intend to use this server in prod remove the next two lines
        app.config["DEBUG"] = True
        app.config['TEMPLATES_AUTO_RELOAD'] = True

        @app.route('/')
        def index():
            """display the homepage"""
            return render_template("index.html")

        @app.route('/<path>')
        def static_file(path:str):
            """Takes in a path and tries to return the template file with that title
            So if a user types in /test then it will try to find a template called test.html

            Parameters
            ----------
            path : str
                The path the user puts in the browser
            """
            try: # look for a page
                if not path.endswith(".html"):
                    if path.endswith(".jpg") or path.endswith("jpeg") or path.endswith("png") or path.endswith("js") or path.endswith("css"):
                        return app.send_static_file(path)
                    return render_template(f"{path}.html")
                else: # If path ends with .html NOTE: This conditional is also a security risk, remove in production
                    return render_template(path)
            except TemplateNotFound:
                try: # Try to look for a 404 page
                    return render_template("404.html")
                except TemplateNotFound: # If no 404 page exists, return a generic 404 page
                    return """<div style='font-size: XXX-large;text-align: center;'>
                    <h1>404 page not found</h1>
                    </br>
                    <button onclick='history.go(-1)'> Click to go back</button>
                    </div>"""
        generate_site(output_folder=temp_dir, preview=True)                          # Generate the site initially
        open_in_browser()                                                               # Open the browser
        server = Server(app)                                                            # Initialize the livereload server
        server.watch(os.path.join(project_folder, 'images/*'), lambda: generate_site(output_folder=temp_dir, preview=True))       # Watch for changes in the images folder
        server.watch(os.path.join(project_folder, '*.yml'), lambda: generate_site(output_folder=temp_dir, preview=True))          # Watch for changes in the yml files
        server.watch(os.path.join(project_folder, 'content/*/*.md'), lambda: generate_site(output_folder=temp_dir, preview=True)) # watch for changes in the markdown files
        server.serve(port=PORT)                                                         # Start the server

if __name__ == '__main__':
    output_dir = 'site' # The folder you want to export your site to in this case /site
    start_server(output_folder=output_dir) 



from flask import render_template, Flask
from ezcv.core import generate_site
from jinja2.exceptions import TemplateNotFound

template_dir = 'site' # The folder you want to export your site to

app = Flask(__name__, static_url_path='', static_folder=template_dir,  template_folder=template_dir)

# Turn these on when you're developing
# app.config["DEBUG"] = True
# app.config['TEMPLATES_AUTO_RELOAD'] = True

@app.route('/')
def index():
    """display the homepage"""
    return render_template("index.html")

@app.route('/<path>')
def static_file(path:str):
    """Takes in a path and tries to return the template file with that title
    So if a user types in /test then it will try to find a template called test.html

    Parameters
    ----------
    path : str
        The path the user puts in the browser
    """
    try:
        return render_template(f"{path}.html")
    except TemplateNotFound:
        try: # Try to look for a 404 page
            return render_template("404.html")
        except TemplateNotFound: # If no 404 page exists, return a generic 404 page
            return """<div style='font-size: XXX-large;text-align: center;'>
            <h1>404 page not found</h1>
            </br>
            <button onclick='history.go(-1)'> Click to go back</button>
            </div>"""

if __name__ == '__main__':
    generate_site(output_folder=template_dir)
    app.run(host='localhost', port=5000)


"""A python-based static site generator for setting up a CV/Resume site

Note
----
If this is your first time using ezcv I would recommend looking at the user docs at https://ezcv.readthedocs.io

Installation
------------
#### From pypi
```pip install ezcv``` or ```sudo pip3 install ezcv```

#### From source
1. ```git clone https://github.com/Descent098/ezcv```
2. ```pip install .``` or ```sudo pip3 install .```

Modules
-------
#### core
The module containing all primary functionality of ezcv including:

- Section parsing
- HTML generation
- Site exporting

#### cli
The module containing all cli functionality of ezcv including:

- Initializing sites
- Generating temporary preview
- Getting lists of themes and/or copying themes

#### Content
This module is for handling all the functionality around content including:

- Getting the list of content directories
- Getting the metadata and contents of content files for sections
- Classes for parsing all extensions

#### Filter

This is a module that provides additional jinja filters to help with theme development

#### Themes

Contains utilities related to theme management, discovery and creation including:

- Section template discovery and creation
- Remote repo management
- Theme discovery & updating

Quickstart
----------
#### Generating a site using all settings defined in "config.yml"
```
from ezcv.core import generate_site

generate_site()
```

#### Generating a site overriding the theme in "config.yml", output directory and specifying to show a preview of the site
```
from ezcv.core import generate_site

generate_site(output_folder="my_site", theme = "aerial", preview = True)
```
"""
__version__ = "0.3.2"


"""Contains all the configuration for the package on pip"""
import setuptools
from ezcv import __version__

def get_content(*filename:str) -> str:
    """ Gets the content of a file or files and returns
    it/them as a string

    Parameters
    ----------
    filename : (str)
        Name of file or set of files to pull content from 
        (comma delimited)
    
    Returns
    -------
    str:
        Content from the file or files
    """
    content = ""
    for file in filename:
        with open(file, "r", encoding="utf-8") as full_description:
            content += full_description.read()
    return content

setuptools.setup(
    name = "ezcv",
    version = __version__,
    author = "Kieran Wood",
    author_email = "kieran@canadiancoding.ca",
    description = "An easy to use personal site generator",
    long_description = get_content("README.md", "CHANGELOG.md"),
    long_description_content_type = "text/markdown",
    project_urls = {
        "User Docs" :      "https://ezcv.readthedocs.io",
        "API Docs"  :      "https://kieranwood.ca/ezcv",
        "Forum":           "https://github.com/Descent098/ezcv/discussions",
        "Source" :         "https://github.com/Descent098/ezcv",
        "Bug Report":      "https://github.com/Descent098/ezcv/issues/new?assignees=Descent098&labels=bug&template=bug_report.md&title=%5BBUG%5D",
        "Feature Request": "https://github.com/Descent098/ezcv/issues/new?labels=enhancement&template=feature_request.md&title=%5BFeature%5D",
        "Roadmap":         "https://github.com/Descent098/ezcv/projects"
    },
    include_package_data = True,
    package_data = {"":["mkdocs.yml", "docs/*", "./themes/*"]},
    packages = setuptools.find_packages(),

    entry_points = { 
            'console_scripts': ['ezcv = ezcv.cli:main']
        },

    install_requires = [
    "docopt",                # Used for argument parsing if you are writing a CLI
    "pyyaml",                # Used for config file parsing
    "jinja2",                # used as middlewear for generating templates
    "markdown",              # Used to parse markdown
    "tqdm",                  # Used to generate progress bars
    "requests",              # Used to download remote themes
    "exifread",              # Used to read exif data from images
    "python-markdown-math",  # Used to render latex math equations
    "colored",               # Used to color terminal output for emphasis
    "pillow",                # Used to do image compression for optimized builds
    "css-html-js-minify",    # Used to minify html, css and JS files for optimized builds
    "md-mermaid",            # Used to render mermaid graphs in markdown
    "livereload",            # Used to auto-reload the site when changes are made
    "flask",                 # Used to create the web server for live reloading
        ],
    extras_require = {
        "dev" : ["mkdocs", # Used to create HTML versions of the markdown docs in the docs directory
                "pdoc3",   # Used to create development docs
                ], 

    },
    classifiers = [
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta"
    ],
)



"""This module is for handling all the functionality around content including:

- Getting the list of content directories
- Getting the metadata and contents of content files for sections
- Classes for parsing all extensions

Classes
-------
Content:
    Base class for other Content types

Markdown:
    Used for parsing markdown content

Image:
    Used for parsing images

Functions
-------
get_content_directories() -> List[str]:
    Gets a list of the existing content directories i.e. ["projects", "education"]

get_section_content() -> List[List[Union[defaultdict, str]]]:
    Takes in a section folder and gets all the content from the files using the Content subclass asigned to the file extension

"""
# Standard Lib Dependencies
import os                                                # Used primarily in path validation
import logging
import datetime
from collections import defaultdict                      # Used to give dicts default args
from dataclasses import dataclass, field                 # Used to improve class performance
from typing import DefaultDict, List, Tuple, Type, Union # Used to provide accurate type hints


# Third Party Dependencies
import exifread            # Used to get metadata of image files
import markdown            # Used to render and read markdown files
from colored import fg     # Used to highlight output with colors, especially errors/warnings


def get_content_directories() -> List[str]:
    """Gets a list of the existing content directories i.e. ["projects", "education"]

    Returns
    -------
    List[str]:
        The list of existing content directories i.e. ["projects", "education"]
    """
    logging.debug("[ezcv get_content_directories()] Getting content directories")
    result:list[str] = []
    for current_path in os.listdir("content"):
        if os.path.isdir(os.path.join("content", current_path)):
            result.append(os.path.join("content", current_path))

    logging.debug(f"[ezcv get_content_directories()] result {result}")
    return result


def get_section_content(section_content_folder: str, examples: bool = False, blog:bool = False) -> List[List[Union[defaultdict, str]]]:
    """Takes in a section folder and gets all the content from the files using the Content subclass asigned to the file extension

    Parameters
    ----------
    section_content_folder : str
        The string representation of the path to the section folder (i.e. 'content/education')

    examples : bool, optional
        Whether or not to render files with example in the name, by default False
    
    blog : bool, optional
        Whether or not the current section is a blog section, by default False

    Returns
    -------
    List[List[Union[defaultdict, str]]]
        A list representing each file in the section of sublists where the metadata is the first element (as a defaultdict), and the HTML is the second (as a string)


    Examples
    --------
    getting the section content of the education section

    ```
    from ezcv.content import get_section_content

    section_content_folder = 'content/education'

    content = get_section_content(section_content_folder)

    print(content[0]) # Prints [defaultdict(<function <lambda> at 0x000001F1B97CE040>, {'title': 'This is the title', 'company': 'This is the company'}), '<p>This is some content</p>']
    ```
    """
    logging.debug(f"[ezcv get_section_content({section_content_folder=}, {examples=}, {blog=})] Getting section content for {section_content_folder}")
    content:List[List[Union[defaultdict, str]]] = []
    extension_handlers:DefaultDict[str, Type] = Content.get_available_extensions()
    logging.debug(f"[ezcv get_section_content()] Beggining file iteration")
    for file_name in os.listdir(section_content_folder):                   # Iterate through the section_content folder and get the content from each file
        if not examples and file_name.startswith("example"):
            continue
        else:
            logging.debug(f"[ezcv get_section_content()] Getting content for {file_name}")
            extension = "." + file_name.lower().split(".")[-1]      # Get the file extension
            if extension_handlers[extension]:                       # Checking if there exists a Content subclass capable of handling the file
                extension_handler = extension_handlers[extension]() # Instantiate the proper extension

                # Get the content and add it to the list
                if not blog:
                    metadata, html = extension_handler.get_content(os.path.join(section_content_folder, file_name))
                    content.append([metadata, html])
                else:
                    metadata, html = extension_handler.get_content(os.path.join(section_content_folder, file_name))
                    # TODO: Add data if not there
                    if not metadata["created"]:
                        metadata["created"] = datetime.datetime.now().strftime("%Y-%m-%d")
                    if not metadata["updated"]:
                        metadata["updated"] = datetime.datetime.now().strftime("%Y-%m-%d")
                    

                    content.append([metadata, html, file_name])
    logging.debug(f"[ezcv get_section_content()] Returning section content {content=}")
    return content


@dataclass
class Content(dict):
    """Base class for other Content types

    Notes
    -----
    - All subclasses are assumed to have implemented:
        - __metadata__(); Returns a defaultdict of metadata
        - __html__(); Returns the HTML to render
        - A list attribute called extensions, for example in markdown it would be
            extensions:List[str] = ['.md', '.markdown', '.mdown', '.mkdn', '.mkd', '.mdwn']

    Methods
    -------
    get_available_extensions() -> DefaultDict[str, Type]:
        Returns a defaultdict of available extensions and corresponding types to render them

    Raises
    ------
    NotImplementedError
        If any of __metadata__(), or __html__() are not implemented in subclass

    Examples
    --------
    ### Get the list of extensions and their handlers
    ```
    extension_handlers = Content.get_available_extensions()
    print(extension_handlers) ''' defaultdict(
        <function Content.get_available_extensions.<locals>.<lambda> at 0x00000240878690D0>, 
        {
        '.md': <class '__main__.Markdown'>, 
        '.markdown': <class '__main__.Markdown'>,
        '.jpg': <class '__main__.Image'>,
        '.png': <class '__main__.Image'>
        }
    )'''

    # Get an extension handler for a specific extension (in thise case .md files)
    print(extension_handlers[".md"]) # <class '__main__.Markdown'>
    ```

    ### Get content of a list of files
    ```
    content = [] # The list that
    filenames = ['file.md', 'image.jpg'] # Gotten from somewhere else
    extension_handlers = Content.get_available_extensions()

    for current_file in filenames:
        extension = '.' + current_file.split('.')[-1]
        if extension_handlers[extension]:
            extension_handler = extension_handlers[extension]()
            metadata, html = extension_handler.get_content(current_file)
            content.append([metadata, html])
    ```
    """


    def get_available_extensions() -> DefaultDict[str, Type]:
        """Returns a defaultdict of extensions and corresponding child types to handle them

        Returns
        -------
        DefaultDict[str, Type]:
            A defaultdict with a str for the extension as a key, and the type as a value
        """
        logging.debug("[ezcv Content.get_available_extensions()] Getting available extensions")
        all_extensions:DefaultDict[str, Type] = defaultdict(lambda:False)
        for current_class in Content.__subclasses__():
            for extension in current_class.extensions:
                all_extensions[extension] = current_class

        logging.debug(f"[ezcv Content.get_available_extensions()] Found extensions {all_extensions=}")
        return all_extensions


    def __metadata__(self):
        """A function to be replaced with the specific implementation of generating metadata defauldict

        Raises
        ------
        NotImplementedError
            If the function is not implemented in the subclass
        """
        raise NotImplementedError


    def get_content(self, file_path:str):
        """Generates the metadata and HTML from a peice of content

        Parameters
        ----------
        file_path : str
            The path to the md file

        Raises
        ------
        NotImplementedError
            If the function is not implemented in the subclass
        """
        raise NotImplementedError


    def __html__(self, file_path:str):
        """A function to be replaced with the specific implementation of generating HTML

        Parameters
        ----------
        file_path: (str)
            The path for the file

        Raises
        ------
        NotImplementedError
            If the function is not implemented in the subclass
        """
        raise NotImplementedError


@dataclass
class Markdown(Content):
    """Used for parsing markdown content

    Notes
    -----
    - Default extensions are 
        - [footnotes](https://python-markdown.github.io/extensions/footnotes/)
        - [tables](https://python-markdown.github.io/extensions/tables/)
        - [toc (Table of contents)](https://python-markdown.github.io/extensions/toc/)
        - [abbr(abbreviations)](https://python-markdown.github.io/extensions/abbreviations/)
        - [def_list(Definition lists)](https://python-markdown.github.io/extensions/definition_lists/)
        - [sane_lists(Sane lists)](https://python-markdown.github.io/extensions/sane_lists/)
        - [meta](https://python-markdown.github.io/extensions/meta)
        - [mdx_math](https://github.com/mitya57/python-markdown-math)
        - [mermaid](https://github.com/oruelle/md_mermaid)

    Examples
    --------
    ```
    from ezcv.content import Markdown

    html, metadata = Markdown().get_content('file_1.md')
    ```
    """
    md:markdown.Markdown = markdown.Markdown(extensions=['meta', 'footnotes', 'tables', 'toc', 'abbr', 'def_list', 'sane_lists', "mdx_math", "md_mermaid"]) # Setup markdown parser with extensions
    extensions:List[str] = (".md", ".markdown", ".mdown", ".mkdn", ".mkd", ".mdwn")


    def __metadata__(self) -> defaultdict:
        """Gets the metadata from the YAML frontmatter of the markdown file

        Notes
        -----
        - This class requires __html__ to be run first, or self.md to be set manually and run self.md.convert(text:str) over the text of the document

        Returns
        -------
        defaultdict
            Returns a defaultdict with the yaml metadata of a peice of content
        """
        logging.debug("[ezcv Markdown.__metadata__()] Getting metadata")
        metadata:defaultdict = defaultdict(lambda:False)
        for key in self.md.Meta: # Create defaultdict out of metadata
            if type(self.md.Meta[key]) == list:
                metadata[key] = self.md.Meta[key][0]
            else:
                metadata[key] = self.md.Meta[key]
        logging.debug(f"[ezcv Markdown.__metadata__()] Returning metadata {metadata=}")
        return metadata


    def __html__(self, file_path:str) -> str:
        """Parses the markdown file and returns a string with the resulting HTML

        Parameters
        ----------
        file_path : str
            The path to the file to generate the HTML for

        Returns
        -------
        str
            The HTML rendered from the markdown file
        """
        logging.debug(f"[ezcv Markdown.__html__()] Getting HTML for {file_path=}")
        with open(f"{file_path}", "r") as mdfile: # Parse markdown file
            text = mdfile.read()
        html = self.md.convert(text) # Convert the markdown content text to hmtl
        logging.debug(f"[ezcv Markdown.__html__()] Returning HTML for {file_path=}")
        return html


    def get_content(self, file_path: str) -> Tuple[defaultdict, str]:
        """Gets the html content of a file, and the metadata of the file

        Parameters
        ----------
        file_path : str
            The path to the file to render

        Returns
        -------
        defaultdict, str
            Returns a defaultdict of the metadata first then the html content as a string

        Raises
        ------
        FileNotFoundError
            If the provided file path does not exist

        Examples
        --------
        Render a file called file_1.md
        ```
        from ezcv.content import Markdown

        metadata, html  = Markdown().get_content('file_1.md')
        ```
        """
        logging.debug(f"[ezcv Markdown.get_content()] Getting content for {file_path=}")
        if not os.path.exists(file_path): # If file doesn't exist
            raise FileNotFoundError(f"{fg(1)} Could not find file: {file_path}{fg(15)}\n")
        html = self.__html__(file_path)
        metadata = self.__metadata__()
        logging.debug(f"[ezcv Markdown.get_content()] Returning content for {file_path=}")
        return metadata, html


@dataclass
class Image(Content):
    """Used for parsing images
    
    Notes
    -----
    - Exif data is NOT available on PNG images (it is only available of jpg and tiff)
    - Since there are so many conditionals it is recommended to use the existing gallery stylesheet

    Examples
    --------
    Render a file called 1.jpg
    ```
    from ezcv.content import Image

    metadata, html = Image().get_content('1.jpg')
    ```
    """
    ignore_exif_data:bool = False
    extensions:List[str] = (".jpg", ".png", ".jpeg", ".gif", ".svg", ".webp", ".apng", ".jfif", ".pjpeg", ".pjp")
    image_paths:List[str] = field(default_factory=lambda: []) # TODO: find way to implement this properly


    def __metadata__(self, filename:str) -> defaultdict:
        """Return the metadata of the file (and exif data if available)

        Parameters
        ----------
        filename : str
            The path to the file to get the metadata from

        Returns
        -------
        defaultdict
            A defaultdict of the keys with the metadata in it
        """
        logging.debug(f"[ezcv Image.__metadata__()] Getting metadata for {filename=}")
        if self.ignore_exif_data:
            return defaultdict(lambda:False)

        elif filename.lower().endswith("jpg"):
            with open(filename ,"rb") as f:
                tags = exifread.process_file(f)
            tags = defaultdict(lambda:False, tags)
            logging.debug(f"[ezcv Image.__metadata__()] Returning EXIF Tags for {filename=} {tags=}")

            return tags

        else:
            logging.debug(f"[ezcv Image.__metadata__()] No EXIF Tags for {filename=}")
            return defaultdict(lambda:False)


    def __html__(self, tags:defaultdict) -> str:
        """Parses the tags of the image and returns HTML for the EXIF data

        Parameters
        ----------
        file_path : str
            The path to the file to generate the HTML for

        Returns
        -------
        str
            The HTML of the EXIF data from the file
        """
        logging.debug(f"[ezcv Image.__html__({tags=})] Getting HTML for {self.image_paths=}")
        html = ""

        # Lens detail
        if tags['EXIF LensModel']:
            html += f"<p class='lens'>{tags['EXIF LensModel']}</p>\n"
        
        # Focal length
        if tags['EXIF FocalLengthIn35mmFilm']:
            if tags['EXIF FocalLengthIn35mmFilm'] != tags['EXIF FocalLength']:
                html += f"<p class='focal-length'>{tags['EXIF FocalLengthIn35mmFilm']}mm (full frame equivalent)</p>\n"
            else:
                html += f"<p class='focal-length'>{tags['EXIF FocalLengthIn35mmFilm']}mm</p>\n"
        else:
            if tags['EXIF FocalLength']:
                html += f"<p class='focal-length'>{tags['EXIF FocalLength']}mm</p>\n"

        # ISO, Shutter speed, Apperture
        if tags['EXIF ISOSpeedRatings']:
            html += f"<p class='iso'>ISO {tags['EXIF ISOSpeedRatings']}</p>\n"
        if tags['EXIF ExposureTime']:
            html += f"<p class='shutter-speed'>{tags['EXIF ExposureTime']} Second(s)</p>\n"
        if tags['EXIF FNumber']:
            from fractions import Fraction
            tags['EXIF FNumber'] = str(float(Fraction(str(tags['EXIF FNumber'])))) # Convert aperture to str i.e. 6.3
            html += f"<p class='aperture'>f{tags['EXIF FNumber']}</p>\n"

        # Camera body details
        if tags['Image Make'] and tags['Image Model']:
            html += f"<p class='camera-type'>{tags['Image Make']} {tags['Image Model']}</p>\n"
        elif tags['Image Make']:
            html += f"<p class='camera-type'>{tags['Image Make']}</p>\n"
        elif tags["Image Model"]:
            html += f"<p class='camera-type'>{tags['Image Model']}</p>\n"
        else:
            ...
        return html


    def get_content(self, file_path: str) -> Tuple[defaultdict, str]:
        """Gets the html content of a file, and the metadata/exif of the file

        Parameters
        ----------
        file_path : str
            The path to the file to render

        Returns
        -------
        str, defaultdict
            Returns the defaultdict of the metadata, and then the html of the exif data

        Raises
        ------
        FileNotFoundError
            If the provided file path does not exist

        Examples
        --------
        Render a file called 1.jpg
        ```
        from ezcv.content import Image

        metadata, html = Image().get_content('1.jpg')
        ```
        """
        logging.debug(f"[ezcv Image.get_content()] Getting content for {file_path=}")
        tags = self.__metadata__(file_path)
        html = self.__html__(tags)
        tags["file_path"] = f"images/gallery/{file_path.split(os.path.sep)[-1]}"
        self.image_paths.append(file_path)
        return tags, html
